#updated 20221227
#oagGetFileFromUser.ps1

function Get-UserSelectedFileFromFileList(
	[parameter(mandatory)]$listFiles, 
	[string]$sOptSelectMessage = '',
	[bool]$bOptShowFullNames = $false
) {
#updated 20221221
# Get filelist index via user input

	[int]$iResult = 0
	[string]$sMsg = ''
	[string]$sFile = ''
	[int]$iIndex = 0

	# build list of files/options
	ForEach ($item in $listFiles) {
		$iIndex++

		if ($bOptShowFullNames -eq $false) {
			$sFile = $($item.name)
		} else {
			$sFile = $($item.FullName)
		}

		if ($iIndex -eq 1) {
			$sMsg = "$iIndex   $sFile"
		} elseif ($iIndex -lt 10) {
			$sMsg = $sMsg + "`r`n$iIndex   $sFile"
		} else {
			$sMsg = $sMsg + "`r`n$iIndex $sFile"
		}
	}

	if ($sOptSelectMessage) {
		$sMsg = $sOptSelectMessage + "`r`n`r`n" + $sMsg
	} else {
		$sMsg = "Enter the number of the option you want.`r`n`r`n" + $sMsg
	}
	$UserResponse = $PlayniteApi.Dialogs.SelectString($sMsg, 'ENTER OPTION NUMBER', '')
	$sMsg = ''

	# check if any input
	if ("$($UserResponse.result)" -ne 'True') {
		# treat as user cancelled
		$__logger.Warn("ENTER OPTION NUMBER - user cancelled")
		return $iResult
	}

	# check input is not empty string
	$UserResponse = "$($UserResponse.SelectedString)"
	if ($UserResponse) {
		if ($UserResponse.TrimEnd() -eq '') {
			$__logger.Warn("ENTER OPTION NUMBER - nullstring")
			return $iResult
		}
	}

	# check input is numeral
	if ($UserResponse -notmatch "^\d+$") {
		$sMsg = "Invalid input made. '$UserResponse'"
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg,'INVALID INPUT')
		return $iResult
	}

	# check input is inside the option range
	$iIndex = $UserResponse
	if ( ($iIndex -gt $($listFiles.count)) -or ($iIndex -lt 1) ) {
		$sMsg = "Invalid input made. Number entered is not a valid choice. '$iIndex'"
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg,'INVALID INPUT')
		return $iResult
	} else {
		$iResult = $iIndex
	}

	return $iResult
}

function Get-UserSelectedPathForExecutable(
	[parameter(mandatory)][string]$sFileNameNoExt,
	[string]$sOptProgramName = ''
) {
#updated 20221227
# Get/Set path via user input pointing to exe

	[string]$sConfigPath = ''
	[string]$sExecutablePath = ''
	[string]$sFileName = $sFileNameNoExt + '.exe'
	[string]$sMsg = ''

	# if not set, reuse exe name
	if (!$sOptProgramName) {
		$sOptProgramName = $sFileNameNoExt
	}

	$sConfigPath = "Config$($sFileNameNoExt)Path.ini"
	$sConfigPath = Join-Path -Path $CurrentExtensionDataPath -ChildPath $sConfigPath
	if ([System.IO.File]::Exists($sConfigPath)) {
		$sExecutablePath = [System.IO.File]::ReadAllLines($sConfigPath)
	} else {
		$sMsg = "After clicking OK, select the $sOptProgramName executable"
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg)
		$sExecutablePath = $PlayniteApi.Dialogs.SelectFile("$sOptProgramName|$sFileName")
		if (!$sExecutablePath) {
			return
		}
		$null = [System.IO.File]::WriteAllLines($sConfigPath, $sExecutablePath)
		$sMsg = "Saved executable path via user input in '$sExecutablePath'."
		$__logger.Info($sMsg)
		$sMsg = "$sOptProgramName path saved"
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg)
	}

	if (![System.IO.File]::Exists($sExecutablePath)) {
		$null = [System.IO.File]::Delete($sConfigPath)
		$sMsg = "Executable not found in user configured path '$sExecutablePath'."
		$__logger.Info($sMsg)
		$sMsg = "$sOptProgramName not found in '$sExecutablePath'. Please run the extension again to configure it to the correct location."
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg)
		$sExecutablePath = ''
	}

	return $sExecutablePath
}