function Get-ThisExtensionName {
	return 'Import Games from Emulators'
}

function GetMainMenuItems() {
	param($getMainMenuItemsArgs)

	[string]$sMenuSection = "@Import Games"
	[string]$sEmulatorName = ''

$sEmulatorName = 'Cemu'

	$menuItem1 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem1.Description = 'Import games from {0}' -f $sEmulatorName
	$menuItem1.FunctionName = "Invoke-ImportGamesFromEmulatorConfig -sEmulatorName {0}" -f $sEmulatorName
	$menuItem1.Icon = "$CurrentExtensionInstallPath\Icons\{0}.ico" -f $sEmulatorName
	$menuItem1.MenuSection = $sMenuSection

$sEmulatorName = 'RetroArch'

	$menuItem2 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem2.Description = 'Import games from {0}' -f $sEmulatorName
	$menuItem2.FunctionName = "Invoke-ImportGamesFromEmulatorConfig -sEmulatorName {0}" -f $sEmulatorName
	$menuItem2.Icon = "$CurrentExtensionInstallPath\Icons\{0}.ico" -f $sEmulatorName
	$menuItem2.MenuSection = $sMenuSection

$sEmulatorName = 'RPCS3'

	$menuItem3 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem3.Description = 'Import games from {0} - Disc Games' -f $sEmulatorName
	$menuItem3.FunctionName = "Invoke-ImportGamesFromEmulatorConfig -sEmulatorName {0}" -f $sEmulatorName
	$menuItem3.Icon = "$CurrentExtensionInstallPath\Icons\{0}.ico" -f $sEmulatorName
	$menuItem3.MenuSection = $sMenuSection

	$menuItem4 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem4.Description = 'Import games from {0} - HDD Games' -f $sEmulatorName
	$menuItem4.FunctionName = "Invoke-ImportGamesFromEmulatorHDD -sEmulatorName {0}" -f $sEmulatorName
	$menuItem4.Icon = "$CurrentExtensionInstallPath\Icons\{0}.ico" -f $sEmulatorName
	$menuItem4.MenuSection = $sMenuSection


<# $sEmulatorName = 'Vita3K'

	$menuItem5 = New-Object Playnite.SDK.Plugins.ScriptMainMenuItem
	$menuItem5.Description = 'Import games from {0} - HDD Games' -f $sEmulatorName
	$menuItem5.FunctionName = "Invoke-ImportGamesFromEmulatorHDD -sEmulatorName {0}" -f $sEmulatorName
	$menuItem5.Icon = "$CurrentExtensionInstallPath\Icons\{0}.ico" -f $sEmulatorName
	$menuItem5.MenuSection = $sMenuSection
 #>

	return $menuItem1, $menuItem2, $menuItem3, $menuItem4
}

function Get-IsDebugMode {
	[bool]$bResult = ([System.IO.File]::Exists("$CurrentExtensionDataPath\DEBUG"))
	return $bResult
}

function Invoke-ImportGamesFromEmulatorConfig([string]$sEmulatorName) {
	Invoke-CallStack -sEmulatorName $sEmulatorName
}

function Invoke-ImportGamesFromEmulatorHDD([string]$sEmulatorName) {
	Invoke-CallStack -sEmulatorName $sEmulatorName -bInstalledVHD $true
}

function Invoke-CallStack(
	[ValidateSet('RPCS3','Cemu','RetroArch','Vita3K')][string]$sEmulatorName,
	[bool]$bInstalledVHD = $false
) {


	# load functions
	. "$CurrentExtensionInstallPath\lib\oagPnDbGetEmu.ps1"
	. "$CurrentExtensionInstallPath\lib\oagImportGame.ps1"
	switch ($sEmulatorName) {
		'RetroArch' {
			. "$CurrentExtensionInstallPath\lib\oag3pRetroArch.ps1"
			. "$CurrentExtensionInstallPath\lib\oagGetFileFromUser.ps1"
		}
	}


	[bool]$bAddGame = $false
	[string]$sDirEmulator = ''
	[string]$sMsg = ''
	[object]$AppList = $null
	[int]$iTotalGamesFound = 0
	[int]$iTotalGamesImported = 0
	[int]$iTotalGamesSkippedAlreadyIn = 0
	[int]$iTotalGamesSkippedInvalidPath = 0
	[int]$iTotalGamesImportFailed = 0
	[string]$fullnameIni = ''
	[string]$sEmulatorProfile = ''
	[string]$sPlatformName = ''
	[bool]$bNameTitleIncludesExtra = $false
	[int]$iMaxRowsDisplay = 45
	[System.Collections.Generic.List[String]]$sArrAdded = @()

	# get emu from Playnite
	$playniteEmulator = $PlayniteApi.Database.Emulators | Where { $_.Name -eq $sEmulatorName } | Select-Object -First 1
	if (!$playniteEmulator) {
		$sMsg = "Couldn't find $sEmulatorName emulator in Playnite. Make sure you have $sEmulatorName configured in Playnite emulators."
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg, $(Get-ThisExtensionName))
		exit
	}

	# verify emu exists
	$sDirEmulator = Get-PathEmulatorFromPlaynite -sEmulatorName $sEmulatorName
	if (!$sDirEmulator) {
		exit
	}

	# set platform name
	switch ($sEmulatorName) {
		'RetroArch' {
			# if multi-platform (e.g. RetroArch, Dolphin), need platform chosen
			$listFiles = Get-ChildItem -Filter *.lpl -LiteralPath (Join-Path -Path $sDirEmulator -ChildPath 'playlists') -File | Sort-Object -Property FullName

			if ($($listFiles.count) -eq 0) {
				$sMsg = "No playlist files found in RetroArch. Build playlists in RetroArch before retrying import."
				$null = $PlayniteApi.Dialogs.ShowMessage($sMsg,$(Get-ThisExtensionName))
				exit
			}

			# user selects playlist
			$iIndex = 0
			$sMsg = "Enter the option number for the list you want to import."
			$Choice = Get-UserSelectedFileFromFileList -listFiles $listFiles -sOptSelectMessage $sMsg
			$sMsg = ''

			if ($($Choice.count) -gt 1) {
				$Choice = $Choice[1]
			}

			if ($Choice -gt 0) {
				$iIndex = $Choice - 1
			}

			if ($iIndex -eq 0) {
				$sMsg = "No list selected."
				$null = $PlayniteApi.Dialogs.ShowMessage($sMsg)
				exit
			}

			$sPlatformName = $($listFiles[$iIndex].name)
			$sPlatformName = $sPlatformName.replace('.lpl','')

			# convert Platform name
			$__logger.Info("RetroArch Platform = '$sPlatformName'")
			$sPlatformName = (RetroArchFriendlyPlatformName -sPlatformName $sPlatformName -bInvert $true)
			$__logger.Info("set Playnite Platform as '$sPlatformName'")
		}
		'Cemu' {$sPlatformName = 'Nintendo Wii U'}
		'RPCS3' {$sPlatformName = 'Sony PlayStation 3'}
		'Vita3K' {$sPlatformName = 'Sony PlayStation Vita'}
	}
	$sMsg = ''


	# verify Playnite has Platform
	$oPlatform = $PlayniteApi.Database.Platforms | Where { $_.Name -eq $sPlatformName } | Select-Object -First 1
	if (!$oPlatform) {
		$sMsg = "Platform '$sPlatformName' does not exist in Playnite."
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg,'ERROR')
		exit
	}

	# get fullname emu config/ini file
	$fullnameIni = switch ($sEmulatorName) {
		'Cemu' { Join-Path -Path $sDirEmulator -ChildPath 'settings.xml' }
		'Dolphin' { Join-Path -Path $sDirEmulator -ChildPath 'User\Cache\gamelist.cache' }
		'RetroArch' { $($listFiles[$iIndex].FullName) }
		'RPCS3' { Join-Path -Path $sDirEmulator -ChildPath 'games.yml' }
		'Vita3K' { Join-Path -Path $env:APPDATA -ChildPath 'Vita3K\Vita3K\ux0\temp\apps.dat' }
	}

	# verify config/ini exists
	if ($fullnameIni) {
		if (![System.IO.File]::Exists($fullnameIni)) {
			$sMsg = " at '$fullnameIni'"
			$fullnameIni = ''
		}
	}
	if (!$fullnameIni) {
		$sMsg = "Could not locate $sEmulatorName config file$sMsg"
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg)
		exit
	}

	# get game collection
	[System.Collections.Generic.List[Object]]$colProcessGames = @()
	if ($bInstalledVHD -eq $true) {

		switch ($sEmulatorName) {
			'RPCS3' {$colProcessGames = Get-GamesCollectionFromRPCS3VirtualHDD -sDirEmulator $sDirEmulator}
			'Vita3K' {$colProcessGames = Get-GamesCollectionFromVita3KVirtualHDD}
		}
	} else {
		$colProcessGames = Get-GamesCollectionFromConfig -sEmulatorName $sEmulatorName -fullnameIni $fullnameIni -oPlatform $oPlatform
	}


	# emulator variations
	switch ($sEmulatorName) {
		'RPCS3' {
			# read game code list
			#$AppList = $(Get-ObjectCompatDatRPCS3 -sDirEmulator $sDirEmulator)
#UNSOLVED have not reverse engineered this file format

			# read game code list
			$AppList = $(Get-ObjectCodeListRPCS3)
		}
		'RetroArch' {
			# verify emulator profile
			$sEmulatorProfile = Get-LplContentEmulatorProfile -filePath $fullnameIni
			if ($sEmulatorProfile) {
				$sEmulatorProfile = $sEmulatorProfile.Split('(')[1]
				$sEmulatorProfile = $sEmulatorProfile.replace(')','')
				if (($playniteEmulator.Profiles | Where-Object { ($_.name -eq $sEmulatorProfile) }).count -ne 1) {
					$sMsg = "Couldn't find matching profile '$sEmulatorProfile' for $sEmulatorName in Playnite."
					$null = $PlayniteApi.Dialogs.ShowMessage($sMsg)
					exit
				}
			} else {
				$sMsg = "There is no default emulator set for $($oPlatform.name) games in RetroArch.`r`n`r`nThis extension can not proceed with the import as it doesn't know which RetroArch emulator you wish to use. Set a default emulator in RetroArch then refresh your playlist before trying this import again."
				$null = $PlayniteApi.Dialogs.ShowMessage($sMsg)
				exit
			}
		}
	}

	# check if user prefers version/region in game title
	$sMsg = ''
	$sMsg = switch ($sEmulatorName) {
		'RetroArch' {
			"Rom names usually include regions, versions and languages.`nDo you want to remove regions from the game names?`n`n(Region detection is incomplete. If in doubt, choose 'No'. This will keep the default title as it appears in RetroArch)`n`nEXAMPLE:`nNo =`n'Super Mario Land (World) (Rev 1)'`n`nYes =`n'Super Mario Land (Rev 1)'"		
		}
	}
	if ($sMsg) {
		$UserResponse = $PlayniteApi.Dialogs.ShowMessage($sMsg, $(Get-ThisExtensionName), 4)
		switch ($sEmulatorName) {
			'RetroArch' {
				$bNameTitleIncludesExtra = ($UserResponse -eq 'No')
			}
		}
	}


	ForEach ($oProfile in $colProcessGames) {

		if ([System.IO.File]::Exists($oProfile.GameDirImage)) {

			# set missing data

			if (!$oProfile.GameDirInstall) {
				$oProfile.GameDirInstall = Split-Path -Path $oProfile.GameDirImage -parent
				if ($sPlatformName -eq 'Nintendo Wii U') {
					if ((Split-Path -Path $oProfile.GameDirInstall -leaf) -eq 'code') {
						$oProfile.GameDirInstall = Split-Path -Path $oProfile.GameDirInstall -parent
					}
				}
			}

			if (!$oProfile.GameName) {
				if ($sEmulatorName -eq 'RPCS3') {
					$oProfile.GameName = Get-GameNameForPS3 -oProfile $oProfile -AppList $AppList
				}
				if ($sEmulatorName -eq 'Vita3K') {
					$oProfile.GameName = Get-GameNameFromRomFileSFO -fullnameEBOOT $oProfile.GameDirImage -bVita $true 
				}
			}

			if (!$oProfile.Region) {
				switch ($sEmulatorName) {
					'RetroArch' {
# very basic region detection from rom filenames (Redump and No-Intro)
						if ($oProfile.GameDirImage -cmatch '(World)') {
							$oProfile.Region = 'World'
						} elseif ($oProfile.GameDirImage -cmatch '(USA, Europe)') {
							$oProfile.Region = 'USA, Europe'
						} elseif ($oProfile.GameDirImage -cmatch '(USA)') {
							$oProfile.Region = 'USA'
						} elseif ($oProfile.GameDirImage -cmatch '(Europe)') {
							$oProfile.Region = 'Europe'
						} elseif ($oProfile.GameDirImage -cmatch '(Japan)') {
							$oProfile.Region = 'Japan'
						}
					}
					'RPCS3' {
						$oProfile.Region = Get-RegionFromTitleIdPS3 -TitleId $oProfile.TitleId
					}
				}
			}

			switch ($sEmulatorName) {
				'RetroArch' {
# very basic detection of version from rom filenames (No-Intro)
# to improve in future release
					# No-Intro convention
					if ($oProfile.GameName -cmatch '(Proto)') {
						$oProfile.Version = 'Proto'
					} elseif ($oProfile.GameName -cmatch '(Demo)') {
						$oProfile.Version = 'Demo'
					} elseif ($oProfile.GameName -cmatch '(Rev [1-B])') {
						$oProfile.Version = $oProfile.GameName.Split('(')[2]
						$oProfile.Version = $($oProfile.Version).replace(')','')
					}


					if ($bNameTitleIncludesExtra -eq $false) {
						if ($oProfile.Region) {
							$oProfile.GameName = $oProfile.GameName.replace(' (' + $oProfile.Region + ')','')
						}
					}
				}
			}
		}


		# confirm OK to add game
		if (![System.IO.File]::Exists($oProfile.GameDirImage)) {
			$sMsg = "skip import of game - invalid path '$oProfile.GameDirImage'"
			$__logger.Warn($sMsg)
			$bAddGame = $false
			$iTotalGamesSkippedInvalidPath++
		} elseif (!$oProfile.GameName) {
			$sMsg = "skip import of game - no game name set '$($oProfile.GameDirInstall)'"
			$__logger.Warn($sMsg)
			$bAddGame = $false
			$iTotalGamesSkippedInvalidPath++
		} else {

			$bAddGame = Confirm-OkayToAddGameExists -sGameName $oProfile.GameName -sGameDirInstall $oProfile.GameDirInstall -sGameDirImage $oProfile.GameDirImage

			if ($bAddGame -eq $true) {

				$bAddGame = Confirm-OkayToAddGameNotImported -sGameName $oProfile.GameName -sGameDirInstall $oProfile.GameDirInstall -sGameDirImage $oProfile.GameDirImage

				if ($bAddGame -eq $false) {
					$iTotalGamesSkippedAlreadyIn++
				}
			} else {
				$iTotalGamesSkippedInvalidPath++
			}
		}

		# proceed with adding game
		if ($bAddGame -eq $true) {
			$bAddGame = Add-NewEmulatedGame -playniteEmulator $playniteEmulator -sGameName $oProfile.GameName -sGameDirInstall $oProfile.GameDirInstall -sGameDirImage $oProfile.GameDirImage -sVersion $oProfile.Version -Platform $oPlatform -sRegion $oProfile.Region -OptionalProfileName $sEmulatorProfile

			if ($bAddGame -eq $true) {
				$iTotalGamesImported++
				$sMsg = "  $($oProfile.GameName) ($($oPlatform.Name))"
				$null = $sArrAdded.Add($sMsg)
			} else {
				$iTotalGamesImportFailed++
			}
		}
	}


	# display results
	$iTotalGamesFound = $colProcessGames.count
	if ($bInstalledVHD -eq $false) {
		$sMsg = 'the config file'
	} else {
		$sMsg = 'the HDD directory'
	}
	$sMsg = "$iTotalGamesFound games detected from {0} for $sEmulatorName`r`n" -f $sMsg
	if ($iTotalGamesFound -gt 0) {
		$sMsg = $sMsg + "`r`n$iTotalGamesImported games imported`r`n"

		if ($iTotalGamesSkippedAlreadyIn -gt 0) {
			$sMsg = $sMsg + "`r`nSkipped $iTotalGamesSkippedAlreadyIn games already in Playnite"
		}
		if ($iTotalGamesSkippedInvalidPath -gt 0) {
			$sMsg = $sMsg + "`r`nSkipped $iTotalGamesSkippedInvalidPath games with invalid path and/or game name"
		}
		if ($iTotalGamesImportFailed -gt 0) {
			$sMsg = $sMsg + "`r`n$iTotalGamesImportFailed import attempts failed (Check Playnite log for details)"
		}

		if ($iTotalGamesImported -gt 0) {
			$sArrAdded = $sArrAdded | Sort-Object $_.name
			[string]$sListNewLined = $($sArrAdded -Join "`r`n").ToString()
		}
	}
	$__logger.Info($sMsg.replace("`r`n",' '))
	if ($iTotalGamesImported -gt 0) {
		$__logger.Info("`r`nAdded New Games:`r`n" + $sListNewLined)
		if ($iTotalGamesImported -le $iMaxRowsDisplay) {
			$sMsg = $sMsg + "`r`n`r`n" + $sListNewLined
		}
	}
	$null = $PlayniteApi.Dialogs.ShowMessage($sMsg,$(Get-ThisExtensionName))
}

function Get-GamesCollectionFromRPCS3VirtualHDD([string]$sDirEmulator) {
	[string]$sMsg = ''
	[string]$pathHddGames = ''
	[System.Collections.Generic.List[Object]]$colProcessGames = @()

	$pathHddGames = Join-Path -Path $sDirEmulator -ChildPath 'dev_hdd0\game\'

	if (![System.IO.Directory]::Exists($pathHddGames)) {
		$sMsg = "Invalid path at $pathHddGames"
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg,'ERROR')
		exit
	}

	# read subdirs into collection
	$listDirs = Get-ChildItem -Filter * -LiteralPath $pathHddGames -Directory -Recurse -Depth 0
	ForEach ( $dir in $listDirs ) {

		if ($dir.name -like 'NP[A-Z][A-B][0-9][0-9][0-9][0-9][0-9]') {

			# reset
			$hashProfile = $null
			$oProfile = $null

			$hashProfile = @{
				TitleId = $dir.name
				GameDirImage = Join-Path -Path $dir.FullName -ChildPath 'USRDIR\EBOOT.BIN'
				GameDirInstall = $dir.FullName
				GameName = ''
				Region = ''
				Version = ''
			}

			$oProfile = [PsCustomObject]$hashProfile
			$null = $colProcessGames.Add($oProfile)
		}
	}

	return $colProcessGames
}

function Get-GamesCollectionFromVita3KVirtualHDD() {
	[string]$sMsg = ''
	[string]$pathHddGames = ''
	[System.Collections.Generic.List[Object]]$colProcessGames = @()

	$pathHddGames = Join-Path -Path $env:APPDATA -ChildPath 'Vita3K\Vita3K\ux0\app\'

	if (![System.IO.Directory]::Exists($pathHddGames)) {
		$sMsg = "Invalid path at $pathHddGames"
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg,'ERROR')
		exit
	}

	# read subdirs into collection
	$listDirs = Get-ChildItem -Filter * -LiteralPath $pathHddGames -Directory -Recurse -Depth 0
	ForEach ( $dir in $listDirs ) {

		if ($dir.name -like 'P[A-Z][A-Z][A-Z][0-9][0-9][0-9][0-9][0-9]') {

			# reset
			$hashProfile = $null
			$oProfile = $null

			$hashProfile = @{
				TitleId = $dir.name
				GameDirImage = Join-Path -Path $dir.FullName -ChildPath 'eboot.bin'
				GameDirInstall = $dir.FullName
				GameName = ''
				Region = ''
				Version = ''
			}

			$oProfile = [PsCustomObject]$hashProfile
			$null = $colProcessGames.Add($oProfile)
		}
	}

	return $colProcessGames
}

function Get-GamesCollectionFromConfig(
	$sEmulatorName,
	$fullnameIni,
	$oPlatform
) {
	[int]$iIndex = 0
	[string]$sMsg = ''
	[System.Collections.Generic.List[Object]]$colProcessGames = @()

	# read emu config/ini file
	try {
		$config = switch ($sEmulatorName) {
			'Cemu' { $(Get-XmlContentCemu -filePath $fullnameIni) }
			'RetroArch' { $(Get-LplContentGames -filePath $fullnameIni) }
			'RPCS3' { $(Get-YmlContent -filePath $fullnameIni) }
		}
	} catch {}

	if (!$config) {
		$sMsg = "No games were detected in the $sEmulatorName file at: '$fullnameIni'`r`n`r`n"
		if ($sEmulatorName -eq 'RetroArch') {
			$sMsg = $sMsg + "Open RetroArch and rescan the playlist for $($oPlatform.name) games before trying this import again."
		} else {
			$sMsg = $sMsg + "This extension works by importing games from configuration files created by an emulator.`r`nIt will not work if you have never added games to your emulator.`r`nOpen $sEmulatorName and set up games in it first before trying this import again."
		}
		$null = $PlayniteApi.Dialogs.ShowMessage($sMsg)
		exit
	}


	ForEach ($key in $config.GetEnumerator()) {

		# reset
		$hashProfile = $null
		$oProfile = $null


		switch ($sEmulatorName) {
			'RPCS3' {
				$hashProfile = @{
					TitleId = $config[$iIndex].Split(': ')[0]
					GameDirImage = ''
					GameDirInstall = $config[$iIndex]
					GameName = ''
					Region = ''
					Version = ''
				}

				$hashProfile.GameDirInstall = $($hashProfile.GameDirInstall).replace("$($hashProfile.TitleId): ",'').replace('/','\')

				if ($hashProfile.GameDirInstall) {
					$hashProfile.GameDirImage = "$($hashProfile.GameDirInstall)" + 'PS3_GAME\USRDIR\EBOOT.BIN'
				}
			}
			'Cemu' {
				$hashProfile = @{
					TitleId = $config.Entry[$iIndex].title_id
					GameDirImage = $config.Entry[$iIndex].path
					GameDirInstall = ''
					GameName = $config.Entry[$iIndex].name
					Region = $config.Entry[$iIndex].region
					Version = ''
				}

				# convert region
				$hashProfile.Region = switch ($hashProfile.Region) {
					'4' {'Europe'}
					'2' {'USA'}
					default {''}
				}
			}
			'RetroArch' {
				$hashProfile = @{
					TitleId = ''
					GameDirImage = $config[$iIndex].path
					GameDirInstall = ''
					GameName = $config[$iIndex].label
					Region = ''
					Version = ''
				}

				if ($hashProfile.GameDirImage -match '.zip#') {
					$hashProfile.GameDirImage = $hashProfile.GameDirImage.Split('#')[0]
				}
			}
		}

		$oProfile = [PsCustomObject]$hashProfile
		$null = $colProcessGames.Add($oProfile)
		$iIndex++
	}

	return $colProcessGames
}

function Get-RegionFromTitleIdPS3($TitleId) {
	[string]$sResult = ''

	if ($TitleId.length -eq 9) {
		$sResult = switch ($TitleId[2]) {
			'A' {'Asia'}
			'C' {'China'}
			'E' {'Europe'}
			'H' {'Hong Kong'}
			'J' {'Japan'}
			'K' {'Korea'}
			'U' {'USA'}
			default {''}
		}
	}
	return $sResult
}

function Get-GameNameForPS3([parameter(mandatory)]$oProfile,$AppList) {
	[string]$sResult = ''
	[string]$sGameVersion = ''

	# SET GAME NAME FOR RPCS3

	# method 1 - read from rom metafile
	$sResult = Get-GameNameFromRomFileSFO -fullnameEBOOT $oProfile.GameDirImage
	if ($sResult) {

		if ($sResult -like '[A-Z][A-Z][A-Z][0-9][0-9][0-9][0-9]*') {
			$sResult = ''
		} else {
			# clean game name

			# a few titles use linefeeds to separate subtitle
			$sResult = $sResult.replace("`r`n",": ").replace("`n",": ")

			# purge registered symbol
			$sResult = $sResult -replace '\xAE', ' '
			$sResult = $sResult -replace '[\xAE]', ' '


			# ASCII only
			#$sResult = $sResult -replace '[^^\x20-\x7E]+', ''
#TO REVIEW purging non ASCII will remove unwanted characters (e.g. tm) but also removes wanted unicode characters
		}
	}

	# method 2 - search code list for match on TitleId
	if (!$sResult) {
		if ($AppList) {

			$sResult = Get-GameNameFromObjectCodeListRPCS3 -AppList $AppList -TitleId $oProfile.TitleId

			if ($sResult) {
				# clean game name
				$sResult = $sResult -replace '\xAE', ' '
				$sResult = $sResult.replace("`n", ' ')
				$sResult = $sResult.replace('Additional Data','')
				$sResult = $sResult.replace('(UPDATE)','').replace('(EU)','')

				if (($oProfile.Region -eq 'Europe') -or ($oProfile.Region -eq 'USA')) {
					$sResult = $sResult -replace '[^^\x20-\x7E\xAE\x99]+', ''
				}

				# detect + remove version from game name
				if ($sResult -like '*[1-9]`.[0-9][0-9]') {
					$sGameVersion = $sResult.Substring($sResult.Length - 4,4)
					$sResult = $sResult.replace($sGameVersion,'').TrimEnd()
					$sGameVersion = ''
				}
			}
		}
	}

	# option 3 - if all other methods failed - read directory folder name
	if (!$sResult) {
		$sResult = $oProfile.GameDirInstall
		$sResult = Split-Path -Path $sResult -leaf

		# remove any instance of game code from name
		if ($oProfile.TitleId) {
			$sResult = ($sResult.replace('[' + $oProfile.TitleId + ']',''))
			$sResult = ($sResult.replace($oProfile.TitleId,''))
		}
	}

	# clean game name
	if ($sResult) {
		$sResult = $sResult.replace('  ', ' ')
		$sResult = $sResult.TrimEnd()
	}

	return $sResult
}

function Get-XmlContentCemu([parameter(mandatory)]$filePath) {


	# load functions
	. "$CurrentExtensionInstallPath\lib\oagXML.ps1"


	[xml]$xml = New-Object XML
	[xml]$xml = Get-Content -Path $filePath

	$oXML = Get-XmlNodes -XmlDocument $xml -NodePath "content.GameCache"
	
	$null = $xml
	return $oXML
}

function Get-LplContentEmulatorProfile($filePath) {
	[object]$ini = $null

	if ([System.IO.File]::Exists($filePath)) {
		$ini = [System.IO.File]::ReadAllLines($filePath) | ConvertFrom-Json
		$ini = $ini.default_core_name
	}
	return $ini
}

function Get-LplContentGames($filePath) {
	[object]$ini = $null

	if ([System.IO.File]::Exists($filePath)) {
		$ini = [System.IO.File]::ReadAllLines($filePath) | ConvertFrom-Json
		$ini = $ini.items
	}
	return $ini
}

function Get-YmlContent($filePath) {
	[object]$ini = $null

	if ([System.IO.File]::Exists($filePath)) {
		$ini = @{}
		$ini = [System.IO.File]::ReadAllLines($filePath) #| ConvertFrom-Xml
	}
	return $ini
}

function Get-ObjectCompatDatRPCS3([parameter(mandatory)][string]$sDirEmulator) {
	[string]$sPath = ''
	[object]$ini = $null

	$sPath = Join-Path -Path $sDirEmulator -ChildPath 'GuiConfigs\compat_database.dat'
	if ([System.IO.File]::Exists($sPath)) {
		$ini = [System.IO.File]::ReadAllLines($sPath) | ConvertFrom-Json
	}
	return $ini
}

function Get-ObjectCodeListRPCS3 {
	[string]$sPath = ''
	[object]$ini = $null

	$sPath = Join-Path -Path $CurrentExtensionInstallPath -ChildPath 'Resources' | Join-Path -ChildPath 'RPCS3codes.json'
	if ([System.IO.File]::Exists($sPath)) {
		$ini = [System.IO.File]::ReadAllLines($sPath) | ConvertFrom-Json
	} else {
		$__logger.Error("No 'RPCS3codes.json' found at '$sPath'")
	}
	return $ini
}

function Get-GameNameFromObjectCodeListRPCS3(
	[parameter(mandatory)][object]$AppList,
	[string]$TitleId
) {
	[string]$sGameName = ''

	if ($TitleId) {
		if ($AppList) {
			ForEach ($code in $AppList) {
				if ($code.ID -eq $TitleId) {
					$sGameName = $code.Name
					break
				}
			}
		}
	}
	return $sGameName
}

function Get-GameNameFromRomFileSFO([string]$fullnameEBOOT,[bool]$bVita = $false) {

	[string]$sGameName = ''
	[string]$fullnameSFO = ''
	[string]$sText = ''
	[int]$iSize = 0
	[int]$iOffsetKeyTbl = 0
	[int]$iOffsetDataTbl = 0
	[int]$iTotalIndexTbl = 0
	[int]$i = 0
	[int]$iOffset = 20

	if ($bVita) {
		$fullnameSFO = $fullnameEBOOT.Replace('eboot.bin','sce_sys\param.sfo')
	} else {
		$fullnameSFO = $fullnameEBOOT.Replace('USRDIR\EBOOT.BIN','PARAM.SFO')
	}

	if (![System.IO.File]::Exists($fullnameSFO)) {
		return
	}

	# $bytes = $null
	# $bytes = [System.IO.File]::ReadAllBytes($fullnameSFO)
	[byte[]]$bytes = Get-Content -LiteralPath $fullnameSFO -Encoding Byte -Raw

	$iSize = $bytes.length

	$iOffsetKeyTbl = $bytes[8] + $(256 * $bytes[9])
	$iOffsetDataTbl = $bytes[12] + $(256 * $bytes[13])
	$iTotalIndexTbl = $bytes[16]	

	# read index table
	[System.Collections.Generic.List[Object]]$colIndexTbl = @()
	For ($i=0; $i -lt $iTotalIndexTbl; $i++) {

		$hashEntry = $null
		$oCustom = $null

		$hashEntry = @{
			offsetKeyTbl = $bytes[$iOffset]
			paramFmt = $bytes[$iOffset + 3]
			paramLen = $bytes[$iOffset + 4] + $($bytes[$iOffset + 5] * 256)
			paramMaxLen = $bytes[$iOffset + 8] + $($bytes[$iOffset + 9] * 256)
			offsetDataTbl = $bytes[$iOffset + 12] + $($bytes[$iOffset + 13] * 256)
		}

		$oCustom = [PsCustomObject]$hashEntry
		$null = $colIndexTbl.Add($oCustom)

		$iOffset = $iOffset + 16
	}

	ForEach ($oCustom in $colIndexTbl) {

		if ($oCustom.paramFmt -eq 2) {
			$sText = ''
			$sText = Get-NullTerminatedTextFromBytes -bytes $bytes -iOffset $($iOffsetKeyTbl + $oCustom.offsetKeyTbl) -iMaxLen 16

			if ($sText -eq 'TITLE') {
				$sGameName = Get-NullTerminatedTextFromBytes -bytes $bytes -iOffset $($iOffsetDataTbl + $oCustom.offsetDataTbl) -iMaxLen $oCustom.paramLen

				# a few titles use linefeeds to separate subtitle
				$sGameName = $sGameName.replace("`r`n",": ").replace("`n",": ")

				break
			}
		}
	}

	return $sGameName
}

function Get-NullTerminatedTextFromBytes(
	[parameter(mandatory)][Array]$bytes,
	[parameter(mandatory)][int]$iOffset,
	[int]$iMaxLen = 127
) {

	[int]$i = 0
	$sBytes = @()

	# fix invalid iMaxLen
	if ($iMaxLen -gt ($bytes.Count - $iOffset)) {
		$iMaxLen = $($bytes.Count - $iOffset)
	}

	For ($i=$iOffset; $i -le ($iOffset + $iMaxLen); $i++) {

		if ($bytes[$i] -eq 0) {
			break
		} else {
			$sBytes += $bytes[$i]
		}
	}

	return [System.Text.Encoding]::UTF8.GetString($sBytes)
}